/****
*     Struc.prg
*
*     Create C-like structures with Clipper.
*
*     (c) 1998 Manos Aspradakis, Greece
*
*/

#include "struc.ch"

STATIC saStrucs := {}
STATIC soTmpStruc

//
CLASS Struc

DATA aMembers   AS {}
DATA StrucName  AS ""

METHOD New( c )                   // Create a new structure
METHOD AddMember( c, x, l )       // Add a member
METHOD EndStruc()                 // Insert new struc and return handle

ENDCLASS

//
METHOD New( cName )
//
LOCAL nPos
Local oStruc, i, aStruc

::StrucName := UPPER( cName )
::aMembers  := {}

// - If the structure has been allready created
// - return only its instance and assign the original values

IF ( ( nPos := ASCAN(saStrucs, {|e| upper(e[ST_NAME]) == ::StrucName }) ) > 0 )
     oStruc := saStrucs[ nPos, ST_INSTANCE ]
     aStruc := saStrucs[ nPos, ST_MEMBERS ]
     FOR i=1 TO LEN( oStruc )
         oStruc[i] := aStruc[ i, MBR_VALUE ]
     NEXT
     RETURN oStruc
ENDIF

RETURN self


//
METHOD AddMember( cName, xType, lInLine )
//
DEFAULT lInline := .F.
     AADD( ::aMembers, { UPPER( cName ), xType, lInLine } )
RETURN Self


//
METHOD EndStruc()
//
LOCAL i, hHandle, oStruc, oTemp, cOrd

/*
** // not really necessary at this point...
** LOCAL nPos
** IF ( ( nPos := ASCAN( saStrucs, {|e| upper(e[ST_NAME]) == ::StrucName }) ) > 0 )
**      RETURN saStrucs[ nPos, ST_INSTANCE ]
** ENDIF
*/

    // Create a new class named ::StrucName

    hHandle := __ClassNew( ::StrucName, LEN( ::aMembers ) )

    FOR i=1 TO LEN( ::aMembers )

        cOrd := PADL( ALLTRIM( STR(i) ), 3, "0")

        IF ValType( ::aMembers[ i, MBR_VALUE ] ) == "B"

             IF ::aMembers[ i, MBR_IS_INLINE ] == .T.  //@@@ is inline code
                  __CLASSADD( hHandle, ;
                              ::aMembers[ i, MBR_NAME ], ;
                              "__BSG" + cOrd )

                  __CLASSADD( hHandle, ;
                              "_" + ::aMembers[ i, MBR_NAME ], ;
                              "__BSG" + cOrd  )

             ELSE                                      //@@@ Is only a block
                  __CLASSADD( hHandle, ;
                              ::aMembers[ i, MBR_NAME ], ;
                              "__IVGET" + cOrd  )

                  __CLASSADD( hHandle, ;
                              "_" + ::aMembers[ i, MBR_NAME ], ;
                              "__IVSET" + cOrd  )
             ENDIF


        ELSE
                  __CLASSADD( hHandle, ;
                              ::aMembers[ i, MBR_NAME ], ;
                              "__IVGET" + cOrd  )

                  __CLASSADD( hHandle, ;
                              "_" + ::aMembers[ i, MBR_NAME ], ;
                              "__IVSET" + cOrd  )

        ENDIF

    NEXT

    oStruc := __ClassIns( hHandle )

    FOR i=1 TO LEN( ::aMembers )
         oStruc[i] := ::aMembers[ i, MBR_VALUE ]
    NEXT

    AADD( saStrucs, { UPPER( ::StrucName ), oStruc, ::aMembers } )

RETURN oStruc


// -[ Friend functions ]-


/****
*     Returns a UNION of structures
*/

//
FUNCTION StrucUnite( cStruc1, cStruc2, cNewStruc )
//
LOCAL i, oS, aNewMembers := {}
LOCAL oStruc1, oStruc2, aStruc1, aStruc2

     cStruc1 := upper( cStruc1 )
     cStruc2 := upper( cStruc2 )

     FOR i=1 TO LEN( saStrucs )

          IF saStrucs[i,ST_NAME] == cStruc1
              oStruc1 := saStrucs[i,ST_INSTANCE]
              aStruc1 := saStrucs[i,ST_MEMBERS]
          ENDIF

          IF saStrucs[i,ST_NAME] == cStruc2
              oStruc2 := saStrucs[i,ST_INSTANCE]
              aStruc2 := saStrucs[i,ST_MEMBERS]
          ENDIF

     NEXT

     oS := Struc():New( cNewStruc )

     FOR i=1 TO LEN( aStruc1 )
          AADD( aNewMembers, { aStruc1[i,1], aStruc1[i,2], aStruc1[i,3] } )
     NEXT

     FOR i=1 TO LEN( aStruc2 )
          AADD( aNewMembers, { aStruc2[i,1], aStruc2[i,2], aStruc2[i,3] } )
     NEXT

     oS:aMembers := aNewMembers

RETURN( oS:EndStruc() )


/****
*     Returns an instance of an existing structure
*/

//
FUNCTION StrucGet( cStruc )
//
Local oTemp, nPos, oStruc, i, aStruc

     cStruc := UPPER( cStruc )

     IF ( ( nPos := ASCAN( saStrucs, {|e| e[ST_NAME] == cStruc }) ) > 0 )

          oStruc := saStrucs[nPos, ST_INSTANCE]
          aStruc := saStrucs[nPos, ST_MEMBERS]
          oTemp  := oOClone( oStruc )

          // Assign initial values to new struc
          FOR i=1 TO LEN( oTemp )
              oTemp[i] := aStruc[i,MBR_VALUE]
          NEXT

     ELSE

         ALERT( "ERROR: Struc "+ cStruc +" Not found")

     ENDIF

RETURN oTemp


/****
*     Creates a deep copy of an array or object
*/

//
STATIC FUNCTION oOClone( oObject, aValues )
//
Local oNew, i

      IF ( Valtype( oObject ) == "O" )

          oNew := __CLASSINS( oObject:ClassH() )
          ASIZE( oNew, LEN( oObject ) )

      ELSE

          oNew := ARRAY( LEN( oObject ) )

      ENDIF

      ACOPY( oObject, oNew )

      FOR i = 1 TO LEN( oNew )

          IF (ValType(oNew[i]) $ "AO")

              oNew[i] := oOClone( oObject[i] )

          ENDIF

      NEXT

RETURN( oNew )

//
FUNCTION StrucNew( cName )
   soTmpStruc := Struc():New( cName )
RETURN( soTmpStruc )


//
PROCEDURE StrucAdd( cMember, xData )
   soTmpStruc:AddMember( cMember, xData )
RETURN

//
FUNCTION StrucEnd()
local oTemp := soTmpStruc:EndStruc()
soTmpStruc  := NIL
RETURN oTemp

